module vga_crtc(pixclk, reset, r, g, b, hs, vs);
	input wire	pixclk;	/* 25Mhz */
	input wire	reset;
	output wire	r, g, b, hs, vs;

/* Pixel pipe size (delay of syncs) - 1. Adjust that based
 * on how long it takes to go from hcnt to rgb output to
 * delay the sync outputs accordingly
 */
localparam pixpipe_size	= 3; /* 4 pixels */

	reg [9:0] hcnt;
	reg [9:0] vcnt;
	reg [pixpipe_size:0] hspipe;
	reg [pixpipe_size:0] vspipe;
	wire hvisible;
	wire vvisible;
	reg [7:0] next_char;
	wire [7:0] font_pixels;
	reg [7:0] out_shift;

	vga_font font(pixclk, next_char, vcnt[3:0], font_pixels);

	/* hcnt/vcnt are 0,0 at the top-left of the display area */
parameter hdisp_start		= 0;	/* 640 display */
parameter hdisp_rbord		= 640;	/*   8 right border */
parameter hdisp_frontp		= 648;	/*   8 front porch */
parameter hdisp_hsync		= 656;	/*  96 hsync (neg) */
parameter hdisp_backp		= 752;	/*  40 back porch */
parameter hdisp_lbord		= 792;	/*   8 left border */
parameter hdisp_end		= 800;

parameter vdisp_start		= 0;	/* 480 display */
parameter vdisp_bbord		= 480;	/*   8 bottom border */
parameter vdisp_frontp		= 488;	/*   2 front porch */
parameter vdisp_vsync		= 490;	/*   2 vsync (neg) */
parameter vdisp_backp		= 492;	/*  25 back porch */
parameter vdisp_tbord		= 517;	/*   8 top border */
parameter vdisp_end		= 525;

	/* H/V counters */
	always @(posedge reset or posedge pixclk) begin
		if (reset) begin
			hcnt <= hdisp_frontp;
			vcnt <= vdisp_frontp;
		end else begin
			if (hcnt == (hdisp_end - 1)) begin
				hcnt <= 0;
			end else begin
				hcnt <= hcnt + 1;
			end
			if (hcnt == (hdisp_frontp - 1)) begin
				if (vcnt == (vdisp_end - 1)) begin
					vcnt <= 0;
				end else begin
					vcnt <= vcnt + 1;
				end
			end
		end
	end

	/* Syncs, slightly delayed to match rgb
	 *
	 * All those < and > compares are going to cost, but it's not a
	 * huge deal on FPGA at the moment. Eventually, I can turn this
	 * into a state machine of decrementers kicking at 0.
	 */
	always @(posedge reset or posedge pixclk) begin
		if (reset) begin
			hspipe <= 'b1;
			vspipe <= 'b1;
		end else begin
			hspipe[0] <= ~(hcnt >= hdisp_hsync && hcnt < hdisp_backp);
			vspipe[0] <= ~(vcnt >= vdisp_vsync && vcnt < vdisp_backp);
			hspipe[pixpipe_size:1] <= hspipe[pixpipe_size-1:0];
			vspipe[pixpipe_size:1] <= vspipe[pixpipe_size-1:0];
		end
	end
	
	assign hs = hspipe[pixpipe_size];
	assign vs = vspipe[pixpipe_size];
	assign hvisible = hcnt < hdisp_rbord;
	assign vvisible = vcnt < vdisp_bbord;
	reg new_pixels;

	/* Random crap display */
	always @(posedge reset or posedge pixclk) begin
		if (reset) begin
			next_char <= 0;
			out_shift <= 0;
			new_pixels <= 0;
		end else begin
			new_pixels <= 0;

			/* We are in the visible zone */
			if (vvisible && hvisible) begin
				/* Pipe step 0, present character to font ROM
				 * Font ROM will return data in 2 cycles. We could
				 * "fix" that (and only wait 1 clock) by using
				 * a bram construct directly I suppose... but for now,
				 * I have to latch the address in vga_font to get xst
				 * to infer an initialized block RAM instead of using
				 * half of my slives for distributed.
				 * Any better idea welcome
				 */
				if (hcnt[2:0] == 3'b000)
					next_char <= hcnt[9:3] + 33 + vcnt[9:4];
				/* Pipe step 1, waiting for font ROM */
				/* Pipe step 2 signal shift reg new font data here */
				if (hcnt[2:0] == 3'b010)
					new_pixels <= 1;
			end

			/* Pipe stage 3
			 * Shift pixels out (we just shift 0's outside of display) */
			if (new_pixels)
				out_shift <= font_pixels;
			else
				out_shift[7:0] <= { out_shift[6:0], 1'b0 };
		end
	end
	assign r = out_shift[7];
	assign g = out_shift[7];
	assign b = out_shift[7];
endmodule
